<?php
/**
 * Soap Client using Microsoft's NTLM Authentication
 * 
 * Copyright (c) 2008 Invest-In-France Agency http://www.invest-in-france.org
 * 
 * Author : Thomas Rabaix
 * 
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 * @link http://rabaix.net/en/articles/2008/03/13/using-soap-php-with-ntlm-authentication
 * @author Thomas Rabaix
 */

require_once("NTLMHttp.php");

/**
 * Soap Client using Microsoft's NTLM Authentication
 * 
 * @link http://rabaix.net/en/articles/2008/03/13/using-soap-php-with-ntlm-authentication
 * @author Thomas Rabaix
 */
class NTLMSoapClient extends SoapClient {
    /**
    * Username for authentication on the exchnage server
    	 * 
    * @var string
    	 */
    protected $user;
    
    /**
     * Password for authentication on the exchnage server
     * 
     * @var string
     */
    protected $password;
    
    /**
     * Server
     *
     * @var string
     *
     */    
    protected $server;
    
    /**
     * Whether or not to validate ssl certificates
     * 
     * @var boolean
     */
    protected $validate = false;
    
    /**
     *
     * @var NTLM_HTTP 
     */
    protected $ntlm = null;
	
    public function __construct($wsdl, $options, $server, $user, $password) {
        $this->server = $server;
        $this->user = $user;
        $this->password = $password;
        parent::__construct($wsdl, $options);
    }
    
	/**
	 * Performs a SOAP request
	 * 
	 * @link http://php.net/manual/en/function.soap-soapclient-dorequest.php
	 * 
	 * @param string $request the xml soap request
	 * @param string $location the url to request
	 * @param string $action the soap action.
	 * @param integer $version the soap version
	 * @param integer $one_way
	 * @return string the xml soap response.
	 */
	public function __doRequest($request, $location, $action, $version,
		$one_way = 0) {
		$headers = array(
			'Method: POST',
			'Connection: Keep-Alive',
			'User-Agent: PHP-SOAP-CURL',
			'Content-Type: text/xml; charset=utf-8',
			'SOAPAction: "'.$action.'"',
		); // end $headers
		$this->__last_request_headers = $headers;
        if(!isset($this->ntlm)) {
            $this->ntlm = new NTLM_HTTP($this->server, $this->user, $this->password);
        }
        $response = $this->ntlm->post($location, $headers, $request);
        if($response['status'] != 200) {
            $msg = "HTTP Status ".$response['status'].": ".$response['statustext'];
            if($response['body'] != "") {
                $msg .= " ({$response['body']})";
            }
            $this->ntlm->closeSocket();
            throw new Exception($msg);
        }
        return $response['body'];
	} // end function __doRequest()
} // end class NTLMSoapClient
